# -*- coding: utf-8 -*-

import logging

from collections import defaultdict
from dateutil.relativedelta import relativedelta
from werkzeug.urls import url_encode

from odoo import api, fields, models, _
from odoo.addons.saas_worker.const import OAUTH_BASE_URL
from odoo.exceptions import AccessError, UserError
from odoo.osv import expression

# from odoo.addons.base_geolocalize.models.res_partner import geo_find, geo_query_address
from odoo.addons.openerp_portal.models import consts

_logger = logging.getLogger(__name__)


class Partner(models.Model):

    _inherit = 'res.partner'

    # Reduced list of name_search fields, as we have been getting mysterious semi-reproductible
    # performance issues when the searched value is 4-5 characters long: PG switches to using a
    # strange query plan based on an index scan on the `display_name` index (created to make the
    # first opening of the list/kanban view fast).
    # Somehow we noticed that removing the `company_registry` term from the searched fields seems
    # to tip the balance in the right direction - and since this field is mostly empty
    # (NULL freq = 99.9%) and redundant with VAT, it seems ok to drop it.
    # The `ref` field is also filled with random nonsense and mostly NULL, so we drop it for good
    # measure.
    # See also PR description for more contextual details.
    _rec_names_search = ['complete_name', 'email', 'vat']

    _check_company_auto = False

    sales_employees = fields.Integer('Sales')
    functional_employees = fields.Integer('Functional')
    technical_employees = fields.Integer('Technical')
    blocking_point = fields.Char('Blocking Point', tracking=True)
    plan = fields.Selection([
        ('plan_to_use', 'Use OpenERP in his company'),
        ('plan_to_sell', 'Offer services to other companies'),
        ('plan_to_test', 'Teacher or student'),
        ('plan_to_newsletter', 'Coming from OpenERP Newsletter')
    ], string='Plan')
    subscription_oauth_scope = fields.Char('Subscription Scope', help='OAuth scope used during subscription', readonly=True)
    enterprise_subscription_ids = fields.One2many('sale.order', string='Enterprise Subscriptions', compute='_compute_enterprise_subscription')
    is_odoo_partner = fields.Boolean('Odoo Partner', compute='_compute_is_odoo_partner', help='Whether the partner has an ongoing Odoo partnership')
    new_enterprise_users_count = fields.Integer('New Enterprise Users', compute='_compute_new_enterprise_user_count', readonly=True, store=False)
    total_enterprise_users_count = fields.Integer('Total Enterprise Users', compute='_compute_total_enterprise_user_count', readonly=True)
    retention_rate = fields.Float('Retention Rate (%)', compute='_compute_retention_rate', digits=(5, 2), readonly=True, store=False)
    certified_experts_count = fields.Integer('Certified Experts', compute='_compute_certified_experts_count', readonly=True, store=True)
    to_review = fields.Boolean(compute='_compute_to_review', search="_search_to_review", readonly=True, store=False)
    # make fleet vehicles reachable from employee via the user's partner record, so we can search
    fleet_contract_ids = fields.One2many('fleet.vehicle.log.contract', 'purchaser_id', 'Fleet Contracts')
    referrer_subscription_ids = fields.One2many('sale.order', 'referrer_id', 'Referrer Subscription')
    referrer_subscription_log_ids = fields.One2many('sale.order.log', 'referrer_id', 'Referrer Subscription Log')
    integrating_subscription_count = fields.Integer("Integrating Subscriptions Count", compute='_compute_integrating_subscription_count')

    def _compute_integrating_subscription_count(self):
        domain = [
            ('referrer_id', 'in', self.ids),
            ('is_subscription', '=', True),
            ('subscription_state', 'in', ['3_progress', '4_paused', '5_renewed', '6_churn'])
        ]
        sub_counts = self.env['sale.order'].sudo().read_group(domain, ['referrer_id'], ['referrer_id'])
        # {partner_id: count}
        mapped_sub_count = {l['referrer_id'][0]: l['referrer_id_count'] for l in sub_counts}

        for partner in self:
            partner.integrating_subscription_count = mapped_sub_count.get(partner.id, 0)

    def _compute_enterprise_subscription(self):
        for partner in self:
            if not isinstance(partner.id, models.NewId):
                partner_ids = self.env['res.partner'].search([('id', 'child_of', partner.commercial_partner_id.id)]).ids
                partner.enterprise_subscription_ids = self.env['sale.order'].search([
                    ('is_subscription', '=', True),
                    '|',
                    ('enterprise_final_customer_id', 'in', partner_ids),
                    ('partner_id', 'in', partner_ids),
                ])
            else:
                partner.enterprise_subscription_ids = self.env['sale.order']

    def button_show_databases(self):
        # search on database_user_ids in sudo since by default no read access
        # uses 2 queries to avoid random choice of index, making it blazing fast...
        emails = [e for e in (self.child_ids.mapped('email') + [self.email]) if e and '@' in e]
        db_ids = self.env['openerp.enterprise.database'].sudo().search([('company_email', 'in', emails)])
        db_ids |= self.env['openerp.enterprise.database'].sudo().search([('database_user_ids.login', 'in', emails)])
        return {
            'name': _('Databases'),
            'view_mode': 'tree,form',
            'res_model': 'openerp.enterprise.database',
            'type': 'ir.actions.act_window',
            'domain': [('id', 'in', db_ids.ids)],
        }

    def _compute_is_odoo_partner(self):
        partnership_tag = self.env.ref('openerp_enterprise.odoo_partnership_tag')

        sub_data = self.env['sale.order'].sudo().read_group([
            ('is_subscription', '=', True),
            ('partner_id', 'in', self.mapped('commercial_partner_id').ids),
            ('subscription_state', 'in', ['3_progress', '4_paused']),
            ('sale_order_template_id.subscription_tag_ids', 'in', [partnership_tag.id]),
        ], ['partner_id'], ['partner_id'])

        sub_count = {row['partner_id'][0]: row['partner_id_count'] for row in sub_data}

        for partner in self:
            partner.is_odoo_partner = bool(sub_count.get(partner.commercial_partner_id.id, 0))

    @api.depends(
        'commercial_partner_id',
        'referrer_subscription_log_ids.include_dashboard',
        'referrer_subscription_log_ids.new_enterprise_user',
        'referrer_subscription_log_ids.referrer_id',
        'referrer_subscription_log_ids.referrer_id.commercial_partner_id',
        'referrer_subscription_log_ids.event_date',
        'referrer_subscription_log_ids.order_id.subscription_state',
        )
    def _compute_new_enterprise_user_count(self):
        """
            Compute the number of new enterprise user during the past year from
            the sale.order.log that belong to the partner.

            Since the referrer on the subscription is always the commercial_partner_id, we can say
            that a sale.order.log belong to a partner
            if the referrer_id = partner.commercial_partner_id

            Except when a partner change it's name, the old partner become a child of the
            new partner. The subscription is updated but the sale.order.log from
            the past remain link to the old partner.
            To make sure to capture all the change done, we are looking for all the change
            that belong directly to the commercial_partner_id but also all the change the belong
            to its children. Finally we sum up those change and assigned the value
            to the main partner and all its children
        """
        # fetch the user count from the last 365 days grouped by partner
        log_data = self.env['sale.order.log'].sudo().read_group(self._get_new_enterprise_user_domain(),
            ['referrer_id', 'new_enterprise_user'], groupby=['referrer_id'])
        qty_per_partner = {d['referrer_id'][0]: d['new_enterprise_user'] for d in log_data}

        # Group all the res.partner that belong to the same commercial_partner_id
        partners_per_commercial = defaultdict(list)
        for p in self.browse(qty_per_partner.keys()):
            partners_per_commercial[p.commercial_partner_id].append(p.id)

        for partner in self:
            # sum of new_user count that belong to the partner
            # with the same commercial_partner_id than the current partner
            partner.new_enterprise_users_count = sum(
                qty_per_partner.get(p, 0) for p in partners_per_commercial.get(partner.commercial_partner_id, []))

    def _compute_total_enterprise_user_count(self):
        subscriptions = self.env['sale.order'].sudo().search([
            ('is_subscription', '=', True),
            ('referrer_id.id', 'in', self.mapped('commercial_partner_id').ids),
            ('subscription_state', 'in', ['3_progress', '4_paused']),
        ])
        qty_per_partner = {}
        for sub in subscriptions:
            for p in sub.partner_id | sub.referrer_id:
                qty_per_partner[p] = qty_per_partner.get(p, 0) + sub.enterprise_users

        for partner in self:
            partner.total_enterprise_users_count = qty_per_partner.get(partner, 0)


    @api.depends(
        'commercial_partner_id',
        'referrer_subscription_ids.referrer_id',
        'referrer_subscription_ids.referrer_id.commercial_partner_id',
        'referrer_subscription_ids.subscription_state',
        'referrer_subscription_ids.end_date',
        'referrer_subscription_ids.close_reason_id.external',
        'referrer_subscription_ids.maintenance_main_subscription_id',
        'referrer_subscription_ids.sale_order_template_id',
    )
    def _compute_retention_rate(self):
        domain = self._get_retention_domain()
        open_data = self.env['sale.order'].sudo().read_group(domain['open'], ['referrer_id'], groupby=['referrer_id'])
        closed_data = self.env['sale.order'].sudo().read_group(domain['close'], ['referrer_id'], groupby=['referrer_id'])

        open_counts = {d['referrer_id'][0]: d['referrer_id_count'] for d in open_data}
        close_counts = {d['referrer_id'][0]: d['referrer_id_count'] for d in closed_data}
        for partner in self:
            try:
                open_count = open_counts.get(partner.commercial_partner_id.id, 0)
                closed_count = close_counts.get(partner.commercial_partner_id.id, 0)
                partner.retention_rate = round(100 * (open_count) / (open_count + closed_count), 2)
            except ZeroDivisionError:  # no subscriptions
                partner.retention_rate = 0

    @api.depends(
        'commercial_partner_id',
        'child_ids.website_tag_ids',
        'website_tag_ids',
    )
    def _compute_certified_experts_count(self):
        odoo_certified = self.env['res.partner'].read_group(self._get_certified_expert_domain(),
            ['commercial_partner_id'], ['commercial_partner_id'])
        certified_by_company = {row['commercial_partner_id'][0]: row['commercial_partner_id_count'] for row in odoo_certified}
        for partner in self:
            partner.certified_experts_count = certified_by_company.get(partner.commercial_partner_id.id, 0)

    @api.depends(
        'new_enterprise_users_count',
        'certified_experts_count',
        'retention_rate',
        'grade_id.target_users',
        'grade_id.target_certif',
        'grade_id.target_retention_rate',
    )
    def _compute_to_review(self):
        for partner in self:
            if partner.grade_id.target_users > 0:
                partner.to_review = not (partner.new_enterprise_users_count >= partner.grade_id.target_users and
                    partner.certified_experts_count >= partner.grade_id.target_certif and
                    partner.retention_rate >= partner.grade_id.target_retention_rate)
            else:
                partner.to_review = False

    def _search_to_review(self, operator, value):
        # Only partner with a grade where targets > 0 can have to_review = true
        # Narrow the records to filter to only those partners
        grades = self.env['res.partner.grade'].search(['|', '|',
            ('target_users', '>', 0),
            ('target_certif', '>', 0),
            ('target_retention_rate', '>', 0),
        ])
        candidates = self.search([('grade_id', 'in', grades.ids)])
        partners_to_review = candidates.filtered(lambda partner: partner.to_review)

        if (operator, value) in [('=', True), ('!=', False)]:
            return [('id', 'in', partners_to_review.ids)]
        else:
            return [('id', 'not in', partners_to_review.ids)]

    def _get_signup_url_for_action(self, url=None, action=None, view_type=None, menu_id=None, res_id=None, model=None):
        if self._context.get('saas_trial_reset'):
            res = dict.fromkeys(self.ids, False)
            for partner in self:
                res[partner.id] = "%s/web/reset_password?%s" % (OAUTH_BASE_URL, url_encode({
                    'db': self._cr.dbname, 'token': partner.signup_token,
                    'redirect': self._context['saas_trial_reset']
                }))
            return res
        else:
            return super(Partner, self)._get_signup_url_for_action(url=url, action=action, view_type=view_type, menu_id=menu_id, res_id=res_id, model=model)

    def _get_new_enterprise_user_domain(self):
        return [
            ('referrer_id', 'child_of', self.mapped('commercial_partner_id').ids),
            ('event_date', '>=', fields.Date.today() - relativedelta(years=1)),
            ('include_dashboard', '=', True),
            ('order_id.subscription_state', 'in', ['3_progress', '4_paused', '5_renewed', '6_churn']),
            '|',
                ('order_id.close_reason_id.external', '=', False),
                ('order_id.close_reason_id', '=', False),
        ]

    def _get_retention_domain(self):
        renew_close_reason_id = self.env.ref('sale_subscription.close_reason_renew').id
        cancel_close_reason_id = self.env.ref('sale_subscription.close_reason_cancel').id
        partnership_tag = self.env.ref('openerp_enterprise.odoo_partnership_tag', raise_if_not_found=False)
        partnership_tag_ids = [partnership_tag.id] if partnership_tag else []
        base_domain = [
            ('referrer_id', 'in', self.mapped('commercial_partner_id').ids),
            ('sale_order_template_id.subscription_tag_ids', 'not in', partnership_tag_ids),
            ('is_subscription', '=', True),
            # Ignore renewed subscription
            '|', ('state', '!=', 'sale'), ('locked', '=', False),
            # Don't count maintenance only subscription
            ('maintenance_main_subscription_id', '=', False),
        ]
        return {
            'open': expression.AND([base_domain, [
                ('subscription_state', 'in', ['3_progress', '4_paused']),
                ('first_contract_date', '>=', fields.Date.today() - relativedelta(years=3))]]
            ),
            'close': expression.AND([base_domain, [
                ('subscription_state', 'in', ['6_churn']),
                ('end_date', '>=', fields.Date.today() - relativedelta(years=3)),
                ('first_contract_date', '>=', fields.Date.today() - relativedelta(years=3)),
                '|',
                ('close_reason_id.external', '=', False),
                ('close_reason_id', '=', False),
                # Ignore sub close because of renewal
                ('close_reason_id', 'not in', (renew_close_reason_id, cancel_close_reason_id)),
            ]]),
            'ignore': expression.AND([base_domain, [
                ('subscription_state', 'in', ['6_churn']),
                ('first_contract_date', '>=', fields.Date.today() - relativedelta(years=3)),
                ('end_date', '>=', fields.Date.today() - relativedelta(years=3)),
                ('close_reason_id.external', '=', True),
            ]]),
        }

    def _get_certified_expert_domain(self):
        odoo_certs = self.env['res.partner.tag'].sudo().search([
            ('id', 'in', consts.ODOO_VALID_CERT_IDS),
        ])
        return [
            ('id', 'child_of', self.mapped('commercial_partner_id').ids),
            ('id', 'not in', self.mapped('commercial_partner_id').ids),
            ('website_tag_ids', 'in', odoo_certs.ids),
        ]

    def _get_retention_details(self):
        Subscription = self.env['sale.order']
        domain = self._get_retention_domain()
        return {
            'subscription_open': Subscription.search(domain['open']),
            'subscription_close': Subscription.search(domain['close']),
            'subscription_ignore': Subscription.search(domain['ignore']),
        }

    def _get_new_enterprise_user_details(self):
        # We want to avoid to translate the Month when we group by event_date
        SubscriptionLog = self.env['sale.order.log'].with_context(lang='en_US')
        domain = self._get_new_enterprise_user_domain()
        total = SubscriptionLog.read_group(domain, ['new_enterprise_user:sum'], groupby=[])
        total_date = SubscriptionLog.read_group(domain, ['new_enterprise_user:sum'], groupby=['event_date'])
        total_sub = SubscriptionLog.read_group(domain, ['new_enterprise_user:sum'], groupby=['subscription_code'])
        details = SubscriptionLog.read_group(domain, ['new_enterprise_user:sum'],
            groupby=['subscription_code', 'event_date'], lazy=False)
        orders = self.env['sale.order'].search([
            ('client_order_ref', 'in', [s['subscription_code'] for s in total_sub]),
            ('state', '=', 'sale'),
        ], order='start_date desc')
        order_per_code = {o.client_order_ref: o for o in orders}

        return {
            'date_keys': [s['event_date'] for s in total_date],
            'subs_keys': [s['subscription_code'] for s in total_sub],
            'total': total,
            'total_per_date': {s['event_date']: s for s in total_date},
            'total_per_sub': {s['subscription_code']: s for s in total_sub},
            'details_per_key': {(s['subscription_code'], s['event_date']): s for s in details},
            'subscriptions': order_per_code,
        }

    def write(self, values):
        if len(self) == 1 and values.get('vat') \
                and (values.get('country_id') or self.country_id.id):
            # attempt to help sales people get the VAT numbers right, and reduce the housekeeping
            # for CHE. The fix method is normally called only for partners created via ecommerce
            # but we want it all the time. (See odoo/odoo@2deb95b852208a8dbfb9640ac8947533c1ad428b)
            vat_country, _ = self._split_vat(values['vat'])
            country_id = (
                self.env['res.country'].search([('code', '=', vat_country.upper())], limit=1).id
                or values.get('country_id')
                or self.country_id.id
            )
            values['vat'] = self.fix_eu_vat_number(country_id, values['vat'])

        # Requested by CHE: forbid changing Partner Name and VAT number once an
        # invoice has been created for this partner. Otherwise we'll get mismatching
        # values for the intracom declaration and all hell breaks loose.
        # => Only employees who know the fiscal implications can do it (e.g. for typos mainly!)
        if ('name' in values or 'vat' in values):
            can_edit = self.env.su or self.env.user.has_group('openerp_enterprise.group_edit_vat')
            for partner in self:
                if ((not can_edit) and
                        (('vat' in values and partner.vat and partner.vat != values['vat']) or
                         ('name' in values and partner.name != values['name'])) and
                        # look for validated invoices
                        not partner.can_edit_vat()):
                    raise AccessError('Changing the name or tax ID of a customer/supplier with '
                                      'validated invoices is not authorized, as it would cause a '
                                      'mismatch in our tax returns. Please create a new one instead.\n\n'
                                      'Contact our helpdesk if you think you are facing '
                                      'an exception to this rule (e.g. small typo in the name)')

        return super(Partner, self).write(values)

    @api.model_create_multi
    def create(self, vals_list):
        for vals in vals_list:
            if not vals.get('phone') or len(vals.get('phone', '')) <= 5:
                vals.pop('phone', None)

            # attempt to help sales people get the VAT numbers right, and reduce the housekeeping
            # for CHE. The fix method is normally called only for partners created via ecommerce
            # but we want it all the time. (See odoo/odoo@2deb95b852208a8dbfb9640ac8947533c1ad428b)
            if vals.get('vat') and vals.get('country_id'):
                vals['vat'] = self.fix_eu_vat_number(vals['country_id'], vals['vat'])

        return super(Partner, self).create(vals_list)

    def action_new_enterprise_user(self):
        return {
            'name': _('New Enterprise Users'),
            'view_mode': 'pivot',
            'res_model': 'sale.order.log.report',
            'type': 'ir.actions.act_window',
            'domain': self._get_new_enterprise_user_domain(),
            'context': {'group_by': 'order_id'},
            'view_id': self.env.ref('openerp_enterprise.sale_order_log_view_pivot').id,
        }

    def action_retention_rate(self):
        domain_dict = self._get_retention_domain()
        domain = expression.OR([domain_dict['open'], domain_dict['close']])
        return {
            'name': _('Subscriptions'),
            'view_mode': 'kanban,tree,form',
            'res_model': 'sale.order',
            'type': 'ir.actions.act_window',
            'domain': domain,
        }

    def action_certified_expert(self):
        return {
            'name': _('Odoo Certified Expert'),
            'view_mode': 'kanban,tree,form',
            'res_model': 'res.partner',
            'type': 'ir.actions.act_window',
            'domain': self._get_certified_expert_domain(),
        }

    def open_partner_dashboard_url(self):
        partnership_tag = self.env.ref('openerp_enterprise.odoo_partnership_tag')

        sub = self.env['sale.order'].search([
            ('partner_id', 'in', self.mapped('commercial_partner_id').ids),
            ('subscription_state', 'in', ['3_progress', '4_paused']),
            ('sale_order_template_id.subscription_tag_ids', 'in', [partnership_tag.id]),
        ])
        if not sub:
            raise UserError(_("No Partner Subscription found. This contact is not an official odoo partner"))

        if len(sub) > 1 and len(set(sub.mapped('client_order_ref'))) == 1:
            sub = self.env['sale.order']._get_most_valid_subscription(sub[0].client_order_ref)

        if len(sub) > 1:
            raise UserError(_("More than 1 Partner Subscription found"))
        return {
            'type': 'ir.actions.act_url',
            'url': '/my/partner/{}'.format(sub.access_token),
            'target': 'self',
        }
